#include <algorithm>
#include <iterator>

#include "pyrainput.h"

void KeyBehaviors::passthrough(unsigned int code) {
	behaviors.at(code - FIRST_KEY).type = KeyBehavior::PASSTHROUGH;
}

void KeyBehaviors::map(unsigned int code, unsigned int result) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::MAPPED;
	b.mapping = result;
}

void KeyBehaviors::altmap(unsigned int code, bool* flag, unsigned int regular, unsigned int alternative) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::ALTMAPPED;
	b.mapping = regular;
	b.alternative = alternative;
	b.flag = flag;
}

void KeyBehaviors::complex(unsigned int code, std::function<void(int)> function) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::COMPLEX;
	b.function = function;
}

void KeyBehaviors::altcomplex(unsigned int code, bool* flag, unsigned int regular, std::function<void(int)> function) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::ALTCOMPLEX;
	b.mapping = regular;
	b.flag = flag;
	b.function = function;
}

void KeyBehaviors::gpmap(unsigned int code, unsigned int gamepad) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::GPMAPPED;
	b.alternative = gamepad;
}

void KeyBehaviors::gpmap2(unsigned int code, unsigned int gamepad, bool* self, LeftRight* pair) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::GPMAP2;
	b.alternative = gamepad;
	b.flag  = self;
	b.lr	= pair;
}

void KeyBehaviors::gphat(unsigned int code, unsigned int gamepad) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::GPHAT;
	b.alternative = gamepad;
}

void KeyBehaviors::handle(unsigned int code, int value) {
	if (code <FIRST_KEY || code >LAST_KEY) return;
	auto& kb = behaviors.at(code - FIRST_KEY);
	switch(kb.type) {
	case KeyBehavior::PASSTHROUGH:
		global.keyboard->send(EV_KEY, code, value);
		break;
	case KeyBehavior::MAPPED:
		global.keyboard->send(EV_KEY, kb.mapping, value);
		break;
	case KeyBehavior::ALTMAPPED:
		if(value==1)
			kb.pressed_as = *kb.flag;
		global.keyboard->send(EV_KEY, kb.pressed_as ? kb.alternative : kb.mapping, value);
		break; 
	case KeyBehavior::COMPLEX:
		kb.function(value);
		break;
	case KeyBehavior::ALTCOMPLEX:
		if(*kb.flag) {
			kb.function(value);
		} else {
			global.keyboard->send(EV_KEY, kb.mapping, value);
		}
		break;
	case KeyBehavior::GPMAPPED:
		if (global.settings.exportKeypad)
			global.keyboard->send(EV_KEY, code, value);
		if (global.settings.exportGamepad) {
			global.gamepad->send(EV_KEY, kb.alternative, value);
			global.gamepad->send(EV_SYN, 0, 0);
		}
		break;
	case KeyBehavior::GPMAP2:
		*kb.flag = (value==1);
		kb.lr->pressed = kb.lr->left || kb.lr->right;
		if (global.settings.exportKeypad)
			global.keyboard->send(EV_KEY, code, value);
		if (global.settings.exportGamepad) {
			global.gamepad->send(EV_KEY, kb.alternative, value);
			global.gamepad->send(EV_SYN, 0, 0);
		}
		break;
	case KeyBehavior::GPHAT:
		global.keyboard->send(EV_KEY, code, value);
		switch (kb.alternative) {
		case BTN_DPAD_UP:
			global.haty = -value;
			break;
		case BTN_DPAD_DOWN:
			global.haty = value;
			break;
		case BTN_DPAD_LEFT:
			global.hatx = -value;
			break;
		case BTN_DPAD_RIGHT:
			global.hatx = value;
			break;
		};
		if (global.settings.exportGamepad) {
			global.gamepad->send(EV_ABS, ABS_HAT0X, global.hatx*65535);
			global.gamepad->send(EV_ABS, ABS_HAT0Y, global.haty*65535);
			global.gamepad->send(EV_SYN, 0, 0);
		}
		break;
	}
}
