#ifndef KEYBEHAVIORS_H
#define KEYBEHAVIORS_H

struct LeftRight {
	bool left	= false;
	bool right	= false;
	bool pressed	= false;
};

static constexpr unsigned int FIRST_KEY = KEY_RESERVED;
static constexpr unsigned int LAST_KEY = KEY_UNKNOWN;

class KeyBehaviors {
public:
	void passthrough(unsigned int code);
	void map(unsigned int code, unsigned int result);
	void gphat(unsigned int code, unsigned int gamepad);
	void gpmap(unsigned int code, unsigned int gamepad);
	void gpmap2(unsigned int code, unsigned int gamepad, bool* self, LeftRight* pair);
	void altmap(unsigned int code, bool* flag, unsigned int regular, unsigned int alternative);
	void altcomplex(unsigned int code, bool* flag, unsigned int regular, std::function<void(int)> function);
	void complex(unsigned int code, std::function<void(int)> function);
	void handle(unsigned int code, int value);

private:
	struct KeyBehavior {
		KeyBehavior() : type(PASSTHROUGH), mapping(0), function() {}
		enum Type  { PASSTHROUGH, MAPPED, ALTMAPPED, COMPLEX, ALTCOMPLEX, GPMAPPED,GPMAP2, GPHAT };
		Type type;
		int mapping;
		int alternative;
		bool pressed_as;
		bool* flag;
		LeftRight* lr;
		std::function<void(int)> function;
	};

	static constexpr unsigned int NUM_KEYS = LAST_KEY - FIRST_KEY + 1;
	std::array<KeyBehavior, NUM_KEYS> behaviors;
};


#endif
